// exbszdat.cpp, v1.0 2013/10/22
// coded by asmodean

// contact: 
//   web:   http://asmodean.reverse.net
//   email: asmodean [at] hush.com
//   irc:   asmodean on efnet (irc.efnet.net)

// This tool extracts GX4Lib (*.dat) archives used by ohXJC [.

#include "as-util.h"
#include "quicklz.h" // QLZ_COMPRESSION_LEVEL=3 QLZ_STREAMING_BUFFER=0

struct DATHDR {
  uint32_t toc_length;
};

struct DATENTRY {
  uint64_t offset;
  uint64_t length;
};

int main(int argc, char** argv) {
  if (argc != 2) {
    fprintf(stderr, "exbszdat v1.0, coded by asmodean\n\n");
    fprintf(stderr, "usage: %s <input.dat>\n", argv[0]);
    return -1;
  }

  string in_filename = argv[1];

  int fd = as::open_or_die(in_filename, O_RDONLY | O_BINARY); 

  DATHDR hdr;
  read(fd, &hdr, sizeof(hdr));

  // Some of them are Unity blobs...
  if (hdr.toc_length & 0xFF000000) {
    fprintf(stderr, "%s: not a supported archive type\n", in_filename.c_str());
    exit(-1);
  }

  uint32_t toc_len  = hdr.toc_length;
  auto     toc_buff = new uint8_t[hdr.toc_length];
  read(fd, toc_buff, toc_len);

  // FileName\0readStartBytePos\0ByteLength
  static const uint8_t ENTRY_START[] = {
    0x46, 0x69, 0x6C, 0x65, 0x4E, 0x61, 0x6D, 0x65, 0x10, 0x72, 0x65, 0x61, 0x64, 0x53, 0x74, 0x61, 
    0x72, 0x74, 0x42, 0x79, 0x74, 0x65, 0x50, 0x6F, 0x73, 0x0A, 0x42, 0x79, 0x74, 0x65, 0x4C, 0x65, 
    0x6E, 0x67, 0x74, 0x68, 
  };

  auto p = toc_buff + toc_len - sizeof(ENTRY_START);

  for (; p >= toc_buff; p--) {
    if (!memcmp(p, ENTRY_START, sizeof(ENTRY_START))) {
      break;
    }
  }

  if (p < toc_buff) {
    fprintf(stderr, "%s: couldn't find TOC\n", in_filename.c_str());
    exit(-1);
  }

  p += 50;

  uint32_t data_base = sizeof(hdr) + toc_len;

  auto end = toc_buff + toc_len;

  while (p < end) {
    uint32_t filename_len = *(uint8_t*) p;
    p += 1;

    string filename = as::convert_utf8(string((char*)p, filename_len));
    p += filename_len;

    uint64_t offset = *(uint64_t*) p;
    p += 8;

    uint32_t len = (uint32_t) *(uint64_t*) p;
    p += 8;

    p += 14;

    auto buff = new uint8_t[len];
    _lseeki64(fd, data_base + offset, SEEK_SET);
    read(fd, buff, len);

    // Laaame.
    if (len < 8 || memcmp(buff, "UnityRaw", 8)) {
      uint32_t out_len   = qlz_size_decompressed((const char*)buff);
      auto     out_buff  = new uint8_t[out_len];
      auto     qlz_state = new qlz_state_decompress;
      memset(qlz_state, 0, sizeof(qlz_state_decompress));
      qlz_decompress((const char*)buff, out_buff, qlz_state);
      delete qlz_state;

      delete [] buff;

      len  = out_len;
      buff = out_buff;
    }

    struct rgba_hdr_t {
      uint32_t width;
      uint32_t height;
      uint32_t width2;
      uint32_t height2;
    };

    auto     rgba_hdr  = (rgba_hdr_t*) buff;
    uint32_t rgba_len  = len  - sizeof(*rgba_hdr);
    auto     rgba_buff = buff + sizeof(*rgba_hdr);

    if (len >= sizeof(*rgba_hdr) &&
        rgba_len == rgba_hdr->width * rgba_hdr->height * 4)
    {
      as::write_bmp(filename + ".bmp",
                    rgba_buff,
                    rgba_len,
                    rgba_hdr->width,
                    rgba_hdr->height,
                    4,
                    as::WRITE_BMP_BGR);
    } else {
      as::write_file(filename, buff, len);
    }

    delete [] buff;
  }

  delete [] toc_buff;

  close(fd);

  return 0;
}
