// merge_spm2.cpp, v1.01 2013//10/23
// coded by asmodean

// contact: 
//   web:   http://asmodean.reverse.net
//   email: asmodean [at] hush.com
//   irc:   asmodean on efnet (irc.efnet.net)

// This tool merges images specified by serialized SPM VER-2.00 objects
// in ohXJC [.  

// Run against a big UnityEngine blob (i.e., resource.assets) or a more
// specific subresource.

#include "as-util.h"

// Apparently all the cool kids love object serialization now days.

uint32_t spm_get_int(uint8_t*& p) {
  uint32_t val = *(uint32_t*) p;
  p += 4;

  return val;
}

string spm_get_string(uint8_t*& p) {
  uint32_t len = spm_get_int(p);

  string s((char*)p, len);
  p += (len + 3) & ~3;

  return s;
}

template<class T>
vector<T>
spm_get_array(uint8_t*& p) {
  uint32_t n = spm_get_int(p);

  vector<T> objects;

  for (uint32_t i = 0; i < n; i++) {
    T object;
    object.load(p);
    objects.push_back(object);
  }

  return objects;
}

struct spm_rect_t {
  void load(uint8_t*& p) {
    x  = spm_get_int(p);
    y  = spm_get_int(p);
    cx = spm_get_int(p);
    cy = spm_get_int(p);
  }

  uint32_t x;
  uint32_t y;
  uint32_t cx;
  uint32_t cy;
};

typedef vector<spm_rect_t> spm_rects_t;

struct spm_chip_data_t {
  void load(uint8_t*& p) {
    image_no    = spm_get_int(p);
    dst_rect.load(p);
    chip_width  = spm_get_int(p);
    chip_height = spm_get_int(p);
    src_rect.load(p);
    unknown1    = spm_get_int(p);
  }

  uint32_t   image_no;
  spm_rect_t dst_rect;
  uint32_t   chip_width;
  uint32_t   chip_height;
  spm_rect_t src_rect;

  uint32_t   unknown1;
};

typedef vector<spm_chip_data_t> spm_chip_datas_t;

struct spm_page_data_t {
  void load(uint8_t*& p) {
    num_chip_data = spm_get_int(p);
    page_width    = spm_get_int(p);
    page_height   = spm_get_int(p);
    page_rect.load(p);
    unknown1      = spm_get_int(p);
    unknown2      = spm_get_int(p);
    hit_rects     = spm_get_array<spm_rect_t>(p);
    chip_datas    = spm_get_array<spm_chip_data_t>(p);
  };

  uint32_t         num_chip_data;
  uint32_t         page_width;
  uint32_t         page_height;
  spm_rect_t       page_rect;

  uint32_t         unknown1;
  uint32_t         unknown2;

  spm_rects_t      hit_rects;
  spm_chip_datas_t chip_datas;
};

typedef vector<spm_page_data_t> spm_page_datas_t;

struct spm_image_data_t {
  void load(uint8_t*& p) {
    image_name = spm_get_string(p);
  }

  string image_name;
};

typedef vector<spm_image_data_t> spm_image_datas_t;

struct spm_int_t {
  void load(uint8_t*& p) {
    value = spm_get_int(p);
  }

  uint32_t value;
};

typedef vector<spm_int_t> spm_ints_t;

struct spm_pat_data_t {
  void load(uint8_t*& p) {
    wait_frame = spm_get_int(p);
    page_no    = spm_get_array<spm_int_t>(p);
  }

  uint32_t   wait_frame;
  spm_ints_t page_no;
};

typedef vector<spm_pat_data_t> spm_pat_datas_t;

struct spm_anim_data_t {
  void load(uint8_t*& p) {
    anim_name               = spm_get_string(p);
    num_pat                 = spm_get_int(p);
    anim_reverse_direction  = spm_get_int(p);
    anim_rotation_direction = spm_get_int(p);
    pat_datas               = spm_get_array<spm_pat_data_t>(p);
  }

  string          anim_name;
  uint32_t        num_pat;
  uint32_t        anim_reverse_direction;
  uint32_t        anim_rotation_direction;
  spm_pat_datas_t pat_datas;
};

typedef vector<spm_anim_data_t> spm_anim_datas_t;

struct spm_data_t {
  void load(uint8_t*& p) {
    version        = spm_get_string(p);
    version_major  = spm_get_int(p);
    version_minor  = spm_get_int(p);

    num_page_data  = spm_get_int(p);
    page_datas     = spm_get_array<spm_page_data_t>(p);

    num_image_data = spm_get_int(p);
    image_datas    = spm_get_array<spm_image_data_t>(p);

    unknown1       = spm_get_int(p);

    num_anim_data  = spm_get_int(p);
    anim_datas     = spm_get_array<spm_anim_data_t>(p);

    path           = spm_get_string(p);
    file           = spm_get_string(p);
  }

  string            version;
  uint32_t          version_major;
  uint32_t          version_minor;

  uint32_t          num_page_data;
  spm_page_datas_t  page_datas;

  uint32_t          num_image_data;
  spm_image_datas_t image_datas;

  uint32_t          unknown1;

  uint32_t          num_anim_data;
  spm_anim_datas_t  anim_datas;

  string            path;
  string            file;
};

int main(int argc, char** argv) {
  if (argc != 2) {
    fprintf(stderr, "merge_spm2 v1.01 by asmodean\n\n");
    fprintf(stderr, "usage: %s <resources.assets>\n\n", argv[0]);
    return -1;
  }

  string in_filename = argv[1];

  uint8_t*   buff = NULL;
  uint32_t   len  = 0;
  as::memory_map_file(in_filename, buff, len);

  as::used_files_t used;

  // There's a simpe serialized TOC, but it seems like a pointless hassle.
  // Scanning by signature is more flexible and doesn't require unpacking
  // all the other garbage resources.

  // (12)"SPM VER-2.00"
  static const uint8_t SPM_START[] = { 0x0C, 0x00, 0x00, 0x00, 0x53, 0x50, 0x4D, 0x20, 0x56, 0x45, 0x52, 0x2D, 0x32, 0x2E, 0x30, 0x30 };

  for (auto p = buff, end = buff + len; p < end - sizeof(SPM_START);) {
    if (!memcmp(p, SPM_START, sizeof(SPM_START))) {
      spm_data_t spm;
      spm.load(p);

      string prefix = "merged/" + as::get_file_prefix(spm.file);

      as::make_path("merged/");

      typedef map<string, as::image_t> cache_t;
      cache_t cache;

      uint32_t page_num = 0;
      for (spm_page_datas_t::iterator page = spm.page_datas.begin();
           page != spm.page_datas.end();
           ++page)
      {
        string out_filename = prefix + as::stringf("+%03d.bmp", page_num++);
        bool   ok           = false;

        as::image_t image(page->page_width, page->page_height, 4);

        for (spm_chip_datas_t::iterator chip = page->chip_datas.begin();
             chip != page->chip_datas.end();
             ++chip)
        {
          auto& filename = spm.image_datas[chip->image_no].image_name;
          auto& part     = cache[filename];

          if (!part.valid()) {
            if (as::image_t::can_open(filename)) {
              part.open(filename);
            } else {
              fprintf(stderr, "%s: can't find part %s\n", out_filename.c_str(), filename.c_str());
            }
          }

          if (part.valid()) {
            ok = true;

            used.add(part.source());

            auto chunk = part.cut(chip->src_rect.x,
                                  chip->src_rect.y,
                                  chip->src_rect.cx,
                                  chip->src_rect.cy,
                                  as::image_t::FLIP_OFFSET);

            chunk.set_offset(chip->dst_rect.x, chip->dst_rect.y);

            image.blend(chunk);
          }
        }

        if (ok) {
          image.trim();
          image.write(out_filename);
        }
      }
    } else {
      p++;
    }
  }

  return 0;
}
