// mergebszvis.cpp, v1.0 2013/10/22
// coded by asmodean

// contact: 
//   web:   http://asmodean.reverse.net
//   email: asmodean [at] hush.com
//   irc:   asmodean on efnet (irc.efnet.net)

// This tool merges the event images from ohXJC [.

#include "as-util.h"

int main(int argc, char** argv) {
  if (argc != 2) {
    fprintf(stderr, "mergebszvis, v1.0 by asmodean\n\n");
    fprintf(stderr, "usage: %s <visual.txt>\n", argv[0]);
    return -1;
  }

  string in_filename(argv[1]);

  FILE* fh = as::open_or_die_file(in_filename);

  while (true) {
    string line = as::read_line(fh, as::READ_LINE_SKIP_JUNK);

    if (line.empty()) {
      break;
    }

    as::split_strings_t fields = as::splitstr(line, ",");

    if (fields.size() == 4) {
      string&  base_filename  = fields[0];
      string&  delta_filename = fields[1];
      uint32_t offset_x       = atol(fields[2].c_str());
      uint32_t offset_y       = atol(fields[3].c_str());

      if (!as::image_t::can_open(base_filename)) {
        fprintf(stderr, "%s: base not found\n", base_filename.c_str());
        continue;
      }

      if (!as::image_t::can_open(delta_filename)) {
        fprintf(stderr, "%s: delta not found\n", delta_filename.c_str());
        continue;
      }
    
      as::image_t base(base_filename);

      as::image_t delta(delta_filename);
      delta.set_offset(offset_x, offset_y);
      delta.set_transparent(as::RGBA::BLACK);

      base.blend_noresize(delta);
      base.write(as::get_file_prefix(delta_filename) + ".bmp");
    } else {
      fprintf(stderr, "%s: malformed input [%s]\n", in_filename.c_str(), line.c_str());
    }
  }

  fclose(fh);

  return 0;
}
